#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Microsoft map tiles provider unit tests.
"""

from geotiler.provider.ms import fromMicrosoftRoad, toMicrosoftRoad, \
    fromMicrosoftAerial, toMicrosoftAerial

import unittest


class MSConversionTestCase(unittest.TestCase):
    """
    Microsfot conversion functions tests.
    """
    def test_from_ms_road(self):
        """
        Test conversion from MS road
        """
        result = fromMicrosoftRoad('0')
        self.assertEqual((0, 0, 1), result)

        result = fromMicrosoftRoad('0230102122203031')
        self.assertEqual((10507, 25322, 16), result)

        result = fromMicrosoftRoad('0230102033330212')
        self.assertEqual((10482, 25333, 16), result)


    def test_to_ms_road(self):
        """
        Test conversion to MS road
        """
        result = toMicrosoftRoad(0, 0, 1)
        self.assertEqual('0', result)

        result = toMicrosoftRoad(10507, 25322, 16)
        self.assertEqual('0230102122203031', result)

        result = toMicrosoftRoad(10482, 25333, 16)
        self.assertEqual('0230102033330212', result)


    def test_from_ms_aerial(self):
        """
        Test conversion from MS aerial
        """
        result = fromMicrosoftAerial('0')
        self.assertEqual((0, 0, 1), result)

        result = fromMicrosoftAerial('0230102122203031')
        self.assertEqual((10507, 25322, 16), result)

        result = fromMicrosoftAerial('0230102033330212')
        self.assertEqual((10482, 25333, 16), result)


    def test_to_ms_aerial(self):
        """
        Test conversion to MS aerial
        """
        result = toMicrosoftAerial(0, 0, 1)
        self.assertEqual('0', result)

        result = toMicrosoftAerial(10507, 25322, 16)
        self.assertEqual('0230102122203031', result)
        result = toMicrosoftAerial(10482, 25333, 16)
        self.assertEqual('0230102033330212', result)


# vim:et sts=4 sw=4:



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
>>> p = BaseProvider('toner', 'png')
>>> p.get_tile_urls((10507, 25322), 16) #doctest: +ELLIPSIS
('http://tile.stamen.com/toner/16/10507/25322.png',)

>>> p = TonerProvider()
>>> p.get_tile_urls((10507, 25322), 16) #doctest: +ELLIPSIS
('http://tile.stamen.com/toner/16/10507/25322.png',)

>>> p = TerrainProvider()
>>> p.get_tile_urls((10507, 25322), 16) #doctest: +ELLIPSIS
('http://tile.stamen.com/terrain/16/10507/25322.jpg',)

>>> p = WatercolorProvider()
>>> p.get_tile_urls((10507, 25322), 16) #doctest: +ELLIPSIS
('http://tile.stamen.com/watercolor/16/10507/25322.jpg',)
"""

from math import pi

from ..geo import MercatorProjection, deriveTransformation
from .base import IMapProvider

URL_FMT = 'http://tile.stamen.com/{style}/{{zoom}}/{{col}}/{{row}}.{ext}'

class BaseProvider(IMapProvider):
    def __init__(self, style, ext):
        # the spherical mercator world tile covers (-π, -π) to (π, π)
        t = deriveTransformation(-pi, pi, 0, 0, pi, pi, 1, 0, -pi, -pi, 0, 1)
        self.projection = MercatorProjection(0, t)

        self.url_fmt = URL_FMT.format(style=style, ext=ext)


    def get_tile_urls(self, tile_coord, zoom):
        column, row = tile_coord
        return (self.url_fmt.format(zoom=zoom, col=column, row=row),)



class TonerProvider(BaseProvider):
    def __init__(self):
        BaseProvider.__init__(self, 'toner', 'png')


class TonerLiteProvider(BaseProvider):
    def __init__(self):
        BaseProvider.__init__(self, 'toner-lite', 'png')


class TerrainProvider(BaseProvider):
    def __init__(self):
        BaseProvider.__init__(self, 'terrain', 'jpg')



class WatercolorProvider(BaseProvider):
    def __init__(self):
        BaseProvider.__init__(self, 'watercolor', 'jpg')


# vim:et sts=4 sw=4:



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

import re
from math import pi, pow

from ..geo import MercatorProjection, deriveTransformation


class IMapProvider(object):
    def __init__(self):
        raise NotImplementedError("Abstract method not implemented by subclass.")


    def get_tile_urls(self, tile_coord, zoom):
        raise NotImplementedError("Abstract method not implemented by subclass.")


    @property
    def tile_width(self):
        return 256


    @property
    def tile_height(self):
        return 256


    def sourceCoordinate(self, tile_coord, zoom):
        wrappedColumn = tile_coord[0] % pow(2, zoom)

        while wrappedColumn < 0:
            wrappedColumn += pow(2, zoom)

        return wrappedColumn, tile_coord[1]



class TemplatedMercatorProvider(IMapProvider):
    """ Convert URI templates into tile URLs, using a tileUrlTemplate identical to:
        http://code.google.com/apis/maps/documentation/overlays.html#Custom_Map_Types
    """
    def __init__(self, template):
        # the spherical mercator world tile covers (-π, -π) to (π, π)
        t = deriveTransformation(-pi, pi, 0, 0, pi, pi, 1, 0, -pi, -pi, 0, 1)
        self.projection = MercatorProjection(0, t)

        self.templates = []

        while template:
            match = re.match(r'^((http|https|file)://\S+?)(,(http|https|file)://\S+)?$', template)
            first = match.group(1)

            if match:
                self.templates.append(first)
                template = template[len(first):].lstrip(',')
            else:
                break


    @property
    def tile_width(self):
        return 256


    @property
    def tile_height(self):
        return 256


    def get_tile_urls(self, tile_coord, zoom):
        x, y = tile_coord
        return [t.format(X=x, Y=y, Z=zoom) for t in self.templates]


# vim: sw=4:et:ai



#
# GeoTiler - library to create maps using tiles from a map provider
#
# Copyright (C) 2014 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This file incorporates work covered by the following copyright and
# permission notice (restored, based on setup.py file from
# https://github.com/stamen/modestmaps-py):
#
#   Copyright (C) 2007-2013 by Michal Migurski and other contributors
#   License: BSD
#

"""
Caching strategies for GeoTiler.
"""

import asyncio
import logging
from functools import partial

from geotiler.tile.io import fetch_tiles

logger = logging.getLogger(__name__)

@asyncio.coroutine
def caching_downloader(get, set, downloader, urls, **kw):
    """
    Create caching map tiles downloader.

    This is asyncio coroutine.

    The code flow is

    - caching downloader gets tiles from cache using URLs
    - the original downloader is used to download missing tiles
    - cache is updated with all existing tiles

    The cache getter function (`get` parameter) should return `None` if
    tile data is not in cache for given URL.

    The collection of tile data is returned for each input URL (or `None`
    if tile data could not be obtained).

    :param get: Function to get a tile from cache.
    :param set: Function to put a tile in cache.
    :param downloader: Original tiles downloader (asyncio coroutine).
    :param urls: Collection of URLs of tiles.
    :param kw: Parameters passed to downloader coroutine.
    """
    data = {u: get(u) for u in urls}
    if __debug__:
        items = (u for u, v in data.items() if v is not None)
        for u in items:
            logger.debug('Cache hit for {}'.format(u))

    # download missing tiles, keep the order of urls
    missing = tuple(u for u in urls if data[u] is None)
    result = yield from downloader(missing, **kw)
    data.update(zip(missing, result))

    # reset cache for new and old tiles
    existing = ((u, t) for u, t in data.items() if t)
    for u, t in existing:
        set(u, t)

    # keep the original order
    return (data[u] for u in urls)


def redis_downloader(client, downloader=None, timeout=3600 * 24 * 7):
    """
    Create downloader using Redis as cache for map tiles.

    :param client: Redis client object.
    :param downloader: Map tiles downloader, use `None` for default downloader.
    :param timeout: Map tile data expiry timeout, default 1 week.
    """
    if downloader is None:
        downloader = fetch_tiles
    set = lambda key, value: client.setex(key, value, timeout)
    return partial(caching_downloader, client.get, set, downloader)


# vim: sw=4:et:ai
